c-----------------------------------------------------------------------
      program calave
c-----------------------------------------------------------------------
c
c --- CALAVE    Version: 7.0.0          Level: 140912              MAIN
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Read a CALPUFF dataset 2.1 or later output file (e.g.,
c               CONC.DAT) and perform either a running average or
c               a block average, for 1 user-specified averaging time
c
c --- UPDATES:
c --- TNG-2.0.0 (140912) to V7.0.0 (140912)
c        1. Update to System 7
c        2. Change CALPUFF dataset TNG-3.0 = 7.0
c           MODIFIED: HD_CFG, HD_PROC
c
c --- v1.1 (120915) to TNG-2.0.0 (140912)
c        1. Add CALPUFF dataset TNG-3.0
c           MODIFIED: HD_CFG, HD_PROC
c                     MODULES - HEADER
c
c --- v1.0 (120702) to v1.1 (120915)
c        1. Change all file pathname variables from C*132 to C*256
c           and make control file line-input strings C*320
c           MODIFIED: (main), CONTROL
c                     MODULES v1.1 - USERINF
c                     UTILS   v1.1 - GETCMD
c
c  EXTERNAL ROUTINES:   GETCMD, CONTROL, ALLOC_1, ALLOC_2, HD_CFG,
c                       HD_PROC, COMP, DEALLOC
c-----------------------------------------------------------------------
c --- Access module data
      use header
      use userinf

      implicit none

      integer :: ictrl,iolst,ioinp,ioout
      parameter(ictrl=10, iolst=11, ioinp=12, ioout=13)


      character(len=12)   :: cversion,clevel
      character(len=256)  :: cfname

      integer :: ione,ierr,ifile

c --- One CALPUFF data file is processed at a time
      data ione/1/

c --- Identify version and level of CALAVE:
      cversion='7.0.0'
      clevel='140912'

c --- Control file name (default or specified on command line)
      cfname='CALAVE.INP'
      call GETCMD(cfname)

c --- Open and process control file
      open(ictrl,file=cfname,status='old')
      call CONTROL(ictrl,iolst,cversion,clevel)
      close(ictrl)

c --- Loop over CALPUFF files (one at a time here)
      do ifile=1,nfiles
         write(*,*)
         write(*,*)'Processing CALPUFF File: '//TRIM(inpfile(ifile))
         write(*,*)'Creating   CALPUFF File: '//TRIM(outfile(ifile))
         write(iolst,*)
         write(iolst,*)'Processing File: '//TRIM(inpfile(ifile))
         write(iolst,*)'Creating   File: '//TRIM(outfile(ifile))

         open(ioinp,file=inpfile(ifile),status='old',
     &              form='unformatted')
         open(ioout,file=outfile(ifile),status='unknown',
     &              form='unformatted')

c ---    Allocate arrays that depend only on the number of
c ---    concurrent CALPUFF files
         call ALLOC_1(ione)

c ---    Process top of file header for model run configuration
         call HD_CFG(iolst,ioinp,ioout,ione,cversion,clevel,ierr)
         if(ierr.NE.0) stop 'HALTED at call to HD_CFG --- see list file'

c ---    Allocate remaining arrays
         call ALLOC_2(ione)
         call ALLOC_3(ione)

c ---    Process rest of file header
         call HD_PROC(ioinp,ioout,ione)

c ---    Process data records
         call COMP(iolst,ioinp,ioout,ierr)
         if(ierr.NE.0) stop 'HALTED at call to COMP --- see list file'

c ---    Close CALPUFF files
         close(ioinp)
         close(ioout)

c ---    Release space for this CALPUFF file
         call DEALLOC

      enddo
c --- End of loop over CALPUFF files

c --- Release space for filenames
      deallocate (inpfile,outfile)

      write(*,*)'Finished Processing All CALPUFF Files'

      stop
      end

c---------------------------
c --- Add utility file
      include 'utils.for'
c---------------------------


c-----------------------------------------------------------------------
      subroutine hd_cfg(iolst,ioinp,ioout,k,cversion,clevel,ierr)
c-----------------------------------------------------------------------
c
c --- CALAVE    Version: 7.0.0          Level: 140912            HD_CFG
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Read first section of a CALPUFF output file header to
c               identify model configuration, and transfer corresponding
c               records to output file (with modifications)
c
c --- UPDATES:
c
c --- TNG-2.0.0 (140912) to v7.0.0 (140912)
c        1. Update CALPUFF dataset TNG-3.0  =  7.0
c
c --- v1.0 (120702) to TNG-2.0.0 (140912)
c        1. Add CALPUFF dataset TNG-3.0
c
c --- INPUT:
c          IOLST - integer - File unit number for output List file
c          IOINP - integer - File unit number for input CALPUFF file
c          IOOUT - integer - File unit number for output CALPUFF file
c              k - integer - Sequential number for input CALPUFF file
c       CVERSION - char*12 - CALAVE version
c         CLEVEL - char*12 - CALAVE level (e.g., date)
c
c --- OUTPUT:
c           IERR - integer - Error flag (0 = No Error)
c
c --- HD_CFG called by: (main)
c --- HD_CFG calls:     SETSTART
c----------------------------------------------------------------------
c --- Access module data
      use userinf
      use header

      implicit none

c --- Arguments
      integer :: iolst,ioinp,ioout,k,ierr
      character(len=12)   :: cversion,clevel

c --- Locals
      integer :: i,iok,idum
      integer :: kbyr,kbjul,kbhr,kbsec,krlg,ireport
      integer :: ksteps
      character(len=132)  :: comment, blank132

c --- Use variable names as listed in the CALPUFF Users Guide

c --- Initializations
      ierr=0
      idum=0
      do i=1,132
         blank132(i:i)=' '
      enddo

c --- Record #1 - File Declaration -- 24 words
      read(ioinp)  dataset(k),dataver(k),datamod(k)
      write(ioout) dataset(k),dataver(k),datamod(k)

c --- QA dataset names
      iok=0
      iok=iok+INDEX(dataset(k),'CONC.DAT')
      iok=iok+INDEX(dataset(k),'DFLX.DAT')
      iok=iok+INDEX(dataset(k),'WFLX.DAT')
      if(iok.EQ.0) then
         write(iolst,*)'ERROR in HD_CFG -- Invalid dataset: '//
     &                  dataset//dataver
         write(iolst,*)'Requires Dataset version 2.1 and later format'
         ierr=1
         return
      endif

c --- Comment records
      read(ioinp)  ncom(k)

c --- Add 3 new comment records to output
      write(ioout) (ncom(k)+3)
         comment='Processed by CALAVE Version '//cversion//
     &           ' Level '//clevel//blank132(60:132)
         write(ioout) comment
         comment='Original CALPUFF output is averaged to nnnn-hour'//
     &           ', nn-minute periods'//blank132(68:132)
         write(comment(40:43),'(i4)') avgpd_hh
         write(comment(51:52),'(i2)') avgpd_mm
         write(ioout) comment
         if(mode.EQ.1) then
            comment='Averaging Method for Output:  RUNNING'//
     &               blank132(38:132)
         else
            comment='Averaging Method for Output:  BLOCK  '//
     &               blank132(38:132)
         endif
         write(ioout) comment

      do i=1,ncom(k)
         comment=blank132
         read(ioinp)  comment
         write(ioout) comment
      enddo

c --- Control variables (record NCOM+3)

      if(dataver(k).EQ.'2.1             ') then
c ---------------------------------------------
         read(ioinp) cmodel(k),ver(k),level(k),ibyr(k),ibjul(k),
     &      ibhr(k),ibsec(k),xbtz(k),irlg(k),iavg(k),nsecdt(k),nxm(k),
     &      nym(k),dxkm(k),dykm(k),nz(k),xorigkm(k),yorigkm(k),
     &      nssta(k),ibcomp(k),iecomp(k),jbcomp(k),jecomp(k),
     &      ibsamp(k),jbsamp(k),iesamp(k),jesamp(k),
     &      meshdn(k),(nsbytype(i,k),i=1,8),
     &      msource(k),ndrec(k),nctrec(k),lsamp(k),
     &      nspout(k),lcompr(k),i2dmet(k),iutmzn(k),feast(k),fnorth(k),
     &      rnlat0(k),relon0(k),xlat1(k),xlat2(k),
     &      pmap(k),utmhem(k),datum(k),daten(k),
     &      clat0(k),clon0(k),clat1(k),clat2(k)
         nsrctype(k)=8
c ---    Rest of source-types are zero;  8 in this file
         do i=nsrctype(k)+1,mxsrctype
            nsbytype(i,k)=0
         enddo
c ---    No receptor groups
         nrgrp(k)=0
c ---    No specie-level units in header
         lspunit(k)=.FALSE.
c ---    ---------------------------------
c ---    Prepare output header information
c ---    ---------------------------------
c ---    Set number of model averages in CALAVE averages
         navgpd=(avgpd_hh*3600+avgpd_mm*60)/(iavg(k)*nsecdt(k))
c ---    Set length of reporting interval in CALAVE output
c ---    and modify start-time and runlength
         kbyr=ibyr(k)
         kbjul=ibjul(k)
         kbhr=ibhr(k)
         kbsec=ibsec(k)
         krlg=irlg(k)
         if(mode.EQ.1) then
c ---       Running average uses original output interval
            ireport=iavg(k)
         elseif(mode.EQ.2) then
c ---       Block average uses CALAVE interval for output
            ireport=navgpd*iavg(k)
         endif
         call SETSTART(iolst,iavg(k),nsecdt(k),
     &                 kbyr,kbjul,kbhr,kbsec,krlg,ksteps,ierr)
         if(ierr.NE.0) return
         write(ioout) cmodel(k),ver(k),level(k),kbyr,kbjul,
     &      kbhr,kbsec,xbtz(k),krlg,ireport,nsecdt(k),nxm(k),
     &      nym(k),dxkm(k),dykm(k),nz(k),xorigkm(k),yorigkm(k),
     &      nssta(k),ibcomp(k),iecomp(k),jbcomp(k),jecomp(k),
     &      ibsamp(k),jbsamp(k),iesamp(k),jesamp(k),
     &      meshdn(k),(nsbytype(i,k),i=1,8),
     &      msource(k),ndrec(k),nctrec(k),lsamp(k),
     &      nspout(k),lcompr(k),i2dmet(k),iutmzn(k),feast(k),fnorth(k),
     &      rnlat0(k),relon0(k),xlat1(k),xlat2(k),
     &      pmap(k),utmhem(k),datum(k),daten(k),
     &      clat0(k),clon0(k),clat1(k),clat2(k)

      elseif(dataver(k).EQ.'2.2             ') then
c ---------------------------------------------
         read(ioinp) cmodel(k),ver(k),level(k),ibyr(k),ibjul(k),
     &      ibhr(k),ibsec(k),axbtz(k),irlg(k),iavg(k),nsecdt(k),nxm(k),
     &      nym(k),dxkm(k),dykm(k),nz(k),xorigkm(k),yorigkm(k),
     &      nssta(k),ibcomp(k),iecomp(k),jbcomp(k),jecomp(k),
     &      ibsamp(k),jbsamp(k),iesamp(k),jesamp(k),
     &      meshdn(k),(nsbytype(i,k),i=1,8),
     &      msource(k),ndrec(k),nctrec(k),lsamp(k),
     &      nspout(k),lcompr(k),i2dmet(k),iutmzn(k),feast(k),fnorth(k),
     &      rnlat0(k),relon0(k),xlat1(k),xlat2(k),
     &      pmap(k),utmhem(k),datum(k),daten(k),
     &      clat0(k),clon0(k),clat1(k),clat2(k)
         nsrctype(k)=8
c ---    Rest of source-types are zero;  8 in this file
         do i=nsrctype(k)+1,mxsrctype
            nsbytype(i,k)=0
         enddo
c ---    No receptor groups
         nrgrp(k)=0
c ---    Specie-level units are in header
         lspunit=.TRUE.
c ---    ---------------------------------
c ---    Prepare output header information
c ---    ---------------------------------
c ---    Set number of model averages in CALAVE averages
         navgpd=(avgpd_hh*3600+avgpd_mm*60)/(iavg(k)*nsecdt(k))
c ---    Set length of reporting interval in CALAVE output
c ---    and modify start-time and runlength
         kbyr=ibyr(k)
         kbjul=ibjul(k)
         kbhr=ibhr(k)
         kbsec=ibsec(k)
         krlg=irlg(k)
         if(mode.EQ.1) then
c ---       Running average uses original output interval
            ireport=iavg(k)
         elseif(mode.EQ.2) then
c ---       Block average uses CALAVE interval for output
            ireport=navgpd*iavg(k)
         endif
         call SETSTART(iolst,iavg(k),nsecdt(k),
     &                 kbyr,kbjul,kbhr,kbsec,krlg,ksteps,ierr)
         if(ierr.NE.0) return
         write(ioout) cmodel(k),ver(k),level(k),kbyr,kbjul,
     &      kbhr,kbsec,axbtz(k),krlg,ireport,nsecdt(k),nxm(k),
     &      nym(k),dxkm(k),dykm(k),nz(k),xorigkm(k),yorigkm(k),
     &      nssta(k),ibcomp(k),iecomp(k),jbcomp(k),jecomp(k),
     &      ibsamp(k),jbsamp(k),iesamp(k),jesamp(k),
     &      meshdn(k),(nsbytype(i,k),i=1,8),
     &      msource(k),ndrec(k),nctrec(k),lsamp(k),
     &      nspout(k),lcompr(k),i2dmet(k),iutmzn(k),feast(k),fnorth(k),
     &      rnlat0(k),relon0(k),xlat1(k),xlat2(k),
     &      pmap(k),utmhem(k),datum(k),daten(k),
     &      clat0(k),clon0(k),clat1(k),clat2(k)

      elseif(dataver(k).EQ.'7.0             ' .OR.
     &       dataver(k).EQ.'TNG-3.0         ') then
c ---------------------------------------------
         read(ioinp) cmodel(k),ver(k),level(k),ibyr(k),ibjul(k),
     &      ibhr(k),ibsec(k),axbtz(k),irlg(k),iavg(k),nsecdt(k),nxm(k),
     &      nym(k),dxkm(k),dykm(k),nz(k),xorigkm(k),yorigkm(k),
     &      nssta(k),ibcomp(k),iecomp(k),jbcomp(k),jecomp(k),
     &      ibsamp(k),jbsamp(k),iesamp(k),jesamp(k),
     &      meshdn(k),nsrctype(k),msource(k),
     &      ndrec(k),nrgrp(k),nctrec(k),lsamp(k),
     &      nspout(k),lcompr(k),i2dmet(k),iutmzn(k),feast(k),fnorth(k),
     &      rnlat0(k),relon0(k),xlat1(k),xlat2(k),
     &      pmap(k),utmhem(k),datum(k),daten(k),
     &      clat0(k),clon0(k),clat1(k),clat2(k)
c ---    Number of each type of source in next record (record NCOM+3a)
         read(ioinp) (nsbytype(i,k),i=1,nsrctype(k))
c ---    Rest of source-types are zero
         do i=nsrctype(k)+1,mxsrctype
            nsbytype(i,k)=0
         enddo
c ---    Specie-level units are in header
         lspunit=.TRUE.
c ---    ---------------------------------
c ---    Prepare output header information
c ---    ---------------------------------
c ---    Set number of model averages in CALAVE averages
         navgpd=(avgpd_hh*3600+avgpd_mm*60)/(iavg(k)*nsecdt(k))
c ---    Set length of reporting interval in CALAVE output
c ---    and modify start-time and runlength
         kbyr=ibyr(k)
         kbjul=ibjul(k)
         kbhr=ibhr(k)
         kbsec=ibsec(k)
         krlg=irlg(k)
         if(mode.EQ.1) then
c ---       Running average uses original output interval
            ireport=iavg(k)
         elseif(mode.EQ.2) then
c ---       Block average uses CALAVE interval for output
            ireport=navgpd*iavg(k)
         endif
         call SETSTART(iolst,iavg(k),nsecdt(k),
     &                 kbyr,kbjul,kbhr,kbsec,krlg,ksteps,ierr)
         if(ierr.NE.0) return
         write(ioout) cmodel(k),ver(k),level(k),kbyr,kbjul,
     &      kbhr,kbsec,axbtz(k),krlg,ireport,nsecdt(k),nxm(k),
     &      nym(k),dxkm(k),dykm(k),nz(k),xorigkm(k),yorigkm(k),
     &      nssta(k),ibcomp(k),iecomp(k),jbcomp(k),jecomp(k),
     &      ibsamp(k),jbsamp(k),iesamp(k),jesamp(k),
     &      meshdn(k),nsrctype(k),msource(k),
     &      ndrec(k),nrgrp(k),nctrec(k),lsamp(k),
     &      nspout(k),lcompr(k),i2dmet(k),iutmzn(k),feast(k),fnorth(k),
     &      rnlat0(k),relon0(k),xlat1(k),xlat2(k),
     &      pmap(k),utmhem(k),datum(k),daten(k),
     &      clat0(k),clon0(k),clat1(k),clat2(k)
         write(ioout) (nsbytype(i,k),i=1,nsrctype(k))
c
c         write(iolst,*)'krlg,ireport,nsecdt(k) = ',
c     &                  krlg,ireport,nsecdt(k)
c
      else
         write(iolst,*)'ERROR in HD_CFG -- Invalid dataset: '//
     &                  dataset//dataver
         write(iolst,*)'Requires Dataset version 2.1 and later format'
         ierr=1
      endif

c --- Derived data
      nsrc(k)=0
      do i=1,mxsrctype
         nsrc(k)=nsrc(k)+nsbytype(i,k)
      enddo
      if(lsamp(k)) then
         ngx(k)=meshdn(k)*(iesamp(k)-ibsamp(k))+1
         ngy(k)=meshdn(k)*(jesamp(k)-jbsamp(k))+1
      else
         ngx(k)=0
         ngy(k)=0
      endif
      ldisc(k)=.FALSE.
      if(ndrec(k).GT.0) ldisc(k)=.TRUE.
      lctsg(k)=.FALSE.
      if(nctrec(k).GT.0) lctsg(k)=.TRUE.

c --- Report selected information to list file
      write(iolst,*)'Number of sources, gridded & discrete receptors'//
     &              ' in output: ',MAX(1,nsrc(k)),(ngx(k)*ngy(k)),
     &               (ndrec(k)+nctrec(k))
      write(iolst,*)'Starting date-time of output '//
     &              '(YYYY,JJJ,HH,SSSS): ',kbyr,kbjul,kbhr,kbsec,
     &               axbtz(k)
      write(iolst,*)'Number of output times, interval (sec): ',ksteps,
     &               (ireport*nsecdt(k))

      return
      end

c-----------------------------------------------------------------------
      subroutine hd_proc(ioinp,ioout,k)
c-----------------------------------------------------------------------
c
c --- CALAVE    Version: 7.0.0          Level: 140912           HD_PROC
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Read remainder of a CALPUFF output file header and 
c               transfer records to output file
c
c --- UPDATES:
c
c --- TNG-2.0.0 (140912) to v7.0.0 (140912)
c        1. Update CALPUFF dataset TNG-3.0  =  7.0
c
c --- v1.0 (120702) to TNG-2.0.0 (140912)
c        1. Add CALPUFF dataset TNG-3.0 (discrete receptor groups)
c
c --- INPUT:
c          IOINP - integer - File unit number for input CALPUFF file
c          IOOUT - integer - File unit number for output CALPUFF file
c              k - integer - Sequential number for output CALPUFF file
c
c --- OUTPUT:
c
c --- HD_PROC called by: (main)
c --- HD_PROC calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use header

      implicit none

c --- Declare arguments
      integer :: ioinp,ioout,k

c --- Locals
      integer :: i,n,itype


c --- Title (record NCOM+4)
      read(ioinp)  (title(i,k),i=1,3)
      write(ioout) (title(i,k),i=1,3)

c --- Specie/level list (record NCOM+5)
      read(ioinp)  (csout(i,k),i=1,nspout(k))
      write(ioout) (csout(i,k),i=1,nspout(k))

c --- Units for each specie/level (record NCOM+5a)
      if(lspunit(k)) then
         read(ioinp)  (csunit(i,k),i=1,nspout(k))
         write(ioout) (csunit(i,k),i=1,nspout(k))
      endif

c --- Non-gridded receptor coordinates (record NCOM+6)
c --- (New in dataset 7.0 / TNG-3.0: htrec, irgrp)
      if(ndrec(k).NE.0) then
         if(nrgrp(k).EQ.0) then
            read(ioinp) (xrec(i,k),i=1,ndrec(k)),
     &                  (yrec(i,k),i=1,ndrec(k)),
     &                  (zrec(i,k),i=1,ndrec(k))
            write(ioout)(xrec(i,k),i=1,ndrec(k)),
     &                  (yrec(i,k),i=1,ndrec(k)),
     &                  (zrec(i,k),i=1,ndrec(k))
         elseif(nrgrp(k).GE.1) then
c ---       (7.0 / TNG-3.0 and up record NCOM+6, 6a)
            read(ioinp) (xrec(i,k),i=1,ndrec(k)),
     &                  (yrec(i,k),i=1,ndrec(k)),
     &                  (zrec(i,k),i=1,ndrec(k)),
     &                  (htrec(i,k),i=1,ndrec(k)),
     &                  (irgrp(i,k),i=1,ndrec(k))
            read(ioinp) (argrpnam(n,k),n=1,nrgrp(k))
            write(ioout)(xrec(i,k),i=1,ndrec(k)),
     &                  (yrec(i,k),i=1,ndrec(k)),
     &                  (zrec(i,k),i=1,ndrec(k)),
     &                  (htrec(i,k),i=1,ndrec(k)),
     &                  (irgrp(i,k),i=1,ndrec(k))
            write(ioout)(argrpnam(n,k),n=1,nrgrp(k))
         endif
      endif

c --- Complex terrain receptor coordinates (record NCOM+7)
      if(nctrec(k).NE.0) then
         read(ioinp)  (xctr(i,k),i=1,nctrec(k)),
     &                (yctr(i,k),i=1,nctrec(k)),
     &                (zctr(i,k),i=1,nctrec(k)),
     &                (ihill(i,k),i=1,nctrec(k))
         write(ioout) (xctr(i,k),i=1,nctrec(k)),
     &                (yctr(i,k),i=1,nctrec(k)),
     &                (zctr(i,k),i=1,nctrec(k)),
     &                (ihill(i,k),i=1,nctrec(k))
      endif

c --- Source names (records NCOM+8 to NCOM+15)
c --- Loop over source types
      do i=1,mxsrctype
         if(nsbytype(i,k).GT.0) then
            read(ioinp)  itype,(csrcnam(n,i,k),n=1,nsbytype(i,k))
            write(ioout) itype,(csrcnam(n,i,k),n=1,nsbytype(i,k))
            if(itype.NE.i) then
c ---          Report problem information and quit
               write(*,*)
               write(*,*)
               write(*,*)'  *  Unexpected source type found, or   *'
               write(*,*)'  *  Unexpected source number found.    *'
               write(*,*)'  *  Source type expected: ',i
               write(*,*)'  *  Source type found   : ',itype
               write(*,*)'  *  Source Names Read:                 *'
               write(*,*) (csrcnam(n,i,k),n=1,nsbytype(i,k))
               stop
            endif
         endif
      enddo

      return
      end

c-----------------------------------------------------------------------
      subroutine setstart(iolst,iavg,nsecdt,kbyr,kbjul,kbhr,kbsec,
     &                    krlg,kavgs,ierr)
c-----------------------------------------------------------------------
c
c --- CALAVE    Version: 7.0.0          Level: 120702          SETSTART
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  QA timing information (control information : model data)
c               and set corresponding start-time and runlength for
c               header of output file
c
c --- INPUT:
c          IOLST - integer - Unit number of output list file
c           IAVG - integer - Number of model timesteps in model output
c         NSECDT - integer - Number of seconds in model timestep
c
c --- OUTPUT:
c          KAVGS - integer - Number of CALAVE averages reported in file
c           IERR - integer - Error flag (0 = No Error)
c
c --- INPUT/OUTPUT:
c           KBYR - integer - Starting year for data in file
c          KBJUL - integer - Starting Julian day for data in file
c           KBHR - integer - Starting hour for data in file
c          KBSEC - integer - Starting second for data in file
c           KRLG - integer - Number of model timesteps in file
c
c --- SETSTART called by: HD_CFG
c --- SETSTART calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use userinf

      implicit none

c --- Declare arguments
      integer :: iolst,iavg,nsecdt,kbyr,kbjul,kbhr,kbsec
      integer :: krlg,kavgs,ierr

c --- Locals
      integer :: ntdiff,iavgsec,nstep
      logical :: ldb

c --- Initialization
      ierr=0
      ldb=.FALSE.

      if(ldb) then
         write(iolst,*)'SETSTART:  File data starts ',kbyr,kbjul,
     &                  kbhr,kbsec
         write(iolst,*)'SETSTART:  File Run length ',krlg
      endif

c --- QA:  requested output averaging time must be multiple of model
c ---      output period
c --- AVGPD_HH,AVGPD_MM is CALAVE averaging period (hours,minutes)
c --- IAVG is period in model output (timesteps)
c --- NAVGPD is number of IAVG steps in CALAVE averaging period
c ---    (computed elsewhere)
c --- IAVG expressed in seconds:
      iavgsec=iavg*nsecdt
      if(((avgpd_hh*3600+avgpd_mm*60)-navgpd*iavgsec).NE.0) then
         write(iolst,*)'ERROR in SETSTART -- user averaging time is'//
     &                 ' not a multiple of the model output in file'
         write(iolst,*)'   Requested Average (hh,mm): ',avgpd_hh,
     &                                                  avgpd_mm
         write(iolst,*)'   Requested Average (sec)  : ',
     &                                (avgpd_hh*3600+avgpd_mm*60)
         write(iolst,*)'   Model Output      (sec)  : ',iavgsec
         ierr=1
         return
      endif

c --- Difference in start times (seconds)
      ntdiff=(start_hour*3600+start_second)-(kbhr*3600+kbsec)
c --- Difference in times (model output steps)
      nstep=ntdiff/iavgsec

c --- QA:  difference must be multiple of model output step
      if((nstep*iavgsec-ntdiff).NE.0) then
         write(iolst,*)'ERROR in SETSTART -- user start time is '//
     &                 'not consistent with times in file'
         write(iolst,*)'   Requested   HHMM : '//start_hhmm
         write(iolst,*)'   Model Start HH,MM: ',kbhr,(kbsec/60)
         write(iolst,*)'   Output Step  SSSS: ',iavgsec
         write(iolst,*)'   Time Diff    SSSS: ',ntdiff
         ierr=1
         return
      endif

c --- Requested start time and averaging period are OK
      kbhr=start_hour
      kbsec=start_second

      if(ntdiff.LT.0) then
c ---    Requested start time is earlier than model start, so start
c ---    the next day at this time
         kbjul=kbjul+1
         if(MOD(kbyr,4).EQ.0 .AND. kbjul.GT.366)then
            kbjul=kbjul-366
            kbyr=kbyr+1
         elseif(kbjul.GT.365)then
            kbjul=kbjul-365
            kbyr=kbyr+1
         endif
      endif

c --- Adjust original runlength for change in start time (timesteps)
      if(ntdiff.NE.0) then
         nstep=ntdiff/nsecdt
         if(nstep.LT.0) nstep=(24*3600/nsecdt)+nstep
         krlg=krlg-nstep
      endif

c --- Adjust runlength for holding complete block-averages
      if(mode.EQ.2) then
         nstep=krlg*nsecdt/(avgpd_hh*3600+avgpd_mm*60)
         krlg=nstep*(avgpd_hh*3600+avgpd_mm*60)/nsecdt
      endif

c --- Compute number of CALAVE averages in file

c      kavgs=krlg
      if(mode.EQ.1) then
c ---    Running averages may have fewer output steps because last
c ---    output step must be a full average
         kavgs=krlg+1-navgpd
      else
c ---    Block averages have integer number of averages in run length
         kavgs=krlg/navgpd
      endif

      if(ldb) then
         write(iolst,*)'SETSTART:  Output starts ',kbyr,kbjul,
     &                  kbhr,kbsec
         write(iolst,*)'SETSTART:  Output Run length ',krlg
         write(iolst,*)'SETSTART:  Output Avg Periods ',kavgs
      endif

      return
      end

c-----------------------------------------------------------------------
      subroutine alloc_1(k)
c-----------------------------------------------------------------------
c
c --- CALAVE    Version: 7.0.0          Level: 140912           ALLOC_1
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Allocate arrays of header information that depend only
c               on the number of CALPUFF files processed
c
c --- UPDATES:
c
c --- v1.0 (120702) to TNG-2.0.0 (140912)
c        1. Add CALPUFF dataset TNG-3.0
c
c --- INPUT:
c              k - integer - Number of CALPUFF files
c
c --- OUTPUT:
c
c --- ALLOC_1 called by: (main)
c --- ALLOC_1 calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use header

      implicit none

c --- Arguments
      integer :: k

c --- Logical arrays (read from file)
      allocate (lcompr(k),lsamp(k))

c --- Logical arrays (derived)
      allocate (ldisc(k),lctsg(k),lspunit(k))

c --- Character arrays (read from file)
      allocate (cmodel(k),ver(k),level(k))
      allocate (dataset(k),dataver(k),datamod(k))
      allocate (title(3,k))
      allocate (pmap(k),datum(k),axbtz(k),daten(k),utmhem(k))
      allocate (clat0(k),clon0(k),clat1(k),clat2(k))

c --- Integer arrays (read from file)
      allocate (ncom(k),ibyr(k),ibjul(k),ibhr(k),ibsec(k),irlg(k))
      allocate (iavg(k),nsecdt(k),nxm(k),nym(k),nz(k),nssta(k))
      allocate (ibcomp(k),iecomp(k),jbcomp(k),jecomp(k))
      allocate (ibsamp(k),jbsamp(k),iesamp(k),jesamp(k))
      allocate (meshdn(k),nsrctype(k),msource(k))
      allocate (ndrec(k),nrgrp(k),nctrec(k),nspout(k))
      allocate (i2dmet(k),iutmzn(k))
      allocate (nsbytype(mxsrctype,k))

c --- Integer arrays (derived)
      allocate (nsrc(k),ngx(k),ngy(k))

c --- Real arrays (read from file)
      allocate (xbtz(k),dxkm(k),dykm(k),xorigkm(k),yorigkm(k))
      allocate (feast(k),fnorth(k))
      allocate (rnlat0(k),relon0(k),xlat1(k),xlat2(k))

      return
      end

c-----------------------------------------------------------------------
      subroutine alloc_2(k)
c-----------------------------------------------------------------------
c
c --- CALAVE    Version: 7.0.0          Level: 140912           ALLOC_2
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Allocate arrays of header information that depend on
c               the number of CALPUFF files processed and other
c               configuration data, and all data arrays
c
c --- UPDATES:
c
c --- v1.0 (120702) to TNG-2.0.0 (140912)
c        1. Add CALPUFF dataset TNG-3.0
c
c --- INPUT:
c              k - integer - Number of CALPUFF files
c
c --- OUTPUT:
c
c --- ALLOC_2 called by: (main)
c --- ALLOC_2 calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use userinf
      use header
      use inpdata

      implicit none

c --- Arguments
      integer :: k

c --- Locals
      integer :: i,n
      integer :: mxspout,mxsrc,mxsrc1,mxsrc2
      integer :: mxdrec,mxctrec,mxnx,mxny,mxrgrp

c --- Get maximum number of species-levels across files
      mxspout=1
      do i=1,k
         mxspout=MAX(mxspout,nspout(i))
      enddo

c --- Get maximum number of discrete receptors across files
      mxdrec=1
      do i=1,k
         mxdrec=MAX(mxdrec,ndrec(i))
      enddo

c --- Get maximum number of gridded receptors across files
      mxnx=1
      mxny=1
      do i=1,k
         mxnx=MAX(mxnx,ngx(i))
         mxny=MAX(mxny,ngy(i))
      enddo

c --- Get maximum number of CTSG receptors across files
      mxctrec=1
      do i=1,k
         mxctrec=MAX(mxctrec,nctrec(i))
      enddo

c --- Get maximum number of each source type across files
      mxsrc=1
      mxsrc1=1
      mxsrc2=1
      do i=1,k
         mxsrc=MAX(mxsrc,nsrc(i))
         do n=1,mxsrctype
            mxsrc1=MAX(mxsrc1,nsbytype(n,i))
         enddo
         if(msource(i).EQ.1) mxsrc2=MAX(mxsrc2,nsrc(i)+1)
      enddo

c --- Get maximum number of receptor groups across files
      mxrgrp=1
      do i=1,k
         mxrgrp=MAX(mxrgrp,nrgrp(i))
      enddo

c --- Character arrays for header
      allocate (csout(mxspout,k),csunit(mxspout,k))
      allocate (csrcnam(mxsrc1,mxsrctype,k))
      allocate (argrpnam(mxrgrp,k))

c --- Integer arrays for header
      allocate (ihill(mxctrec,k))
      allocate (irgrp(mxdrec,k))
      
c --- Real arrays for header
      allocate (xrec(mxdrec,k),yrec(mxdrec,k))
      allocate (zrec(mxdrec,k),htrec(mxdrec,k))
      allocate (xctr(mxctrec,k),yctr(mxctrec,k),zctr(mxctrec,k))


c --- Character arrays for data
      allocate (sname(mxsrc,k))
      allocate (cspecg(mxspout,k),cspecd(mxspout,k),cspecct(mxspout,k))

c --- Integer arrays for data
      allocate (istype(mxsrc,k),isnum(mxsrc,k))
      allocate (nyrb(navgpd,k),njulb(navgpd,k))
      allocate (nhrb(navgpd,k),nsecb(navgpd,k))
      allocate (nyre(navgpd,k),njule(navgpd,k))
      allocate (nhre(navgpd,k),nsece(navgpd,k))

c --- Real arrays for data
      allocate (sxkm(mxsrc,k),sykm(mxsrc,k))
      allocate (datag(mxnx,mxny,mxspout,mxsrc2,navgpd,k))
      allocate (datad(mxdrec,mxspout,mxsrc2,navgpd,k))
      allocate (datact(mxctrec,mxspout,mxsrc2,navgpd,k))

c --- Real work arrays
      allocate (workg(mxnx,mxny))
      allocate (workd(mxdrec))
      allocate (workct(mxctrec))

      return
      end

c-----------------------------------------------------------------------
      subroutine alloc_3(k)
c-----------------------------------------------------------------------
c
c --- CALAVE    Version: 7.0.0          Level: 120702           ALLOC_3
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Allocate arrays for constructing averages
c
c --- INPUT:
c              k - integer - Number of the CALPUFF array to use
c
c --- OUTPUT:
c
c --- ALLOC_3 called by: (main)
c --- ALLOC_3 calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use header
      use avgdata

      implicit none

c --- Arguments
      integer :: k

c --- Locals
      integer :: one

      data one/1/

c --- Real arrays for data
      if(msource(k).EQ.1) then
c ---    Individual source contributions plus total
         allocate (avgdatag(ngx(k),ngy(k),nspout(k),nsrc(k)+1))
         allocate (avgdatad(ndrec(k),nspout(k),nsrc(k)+1))
         allocate (avgdatact(nctrec(k),nspout(k),nsrc(k)+1))
      else
c ---    Just total over all sources
         allocate (avgdatag(ngx(k),ngy(k),nspout(k),one))
         allocate (avgdatad(ndrec(k),nspout(k),one))
         allocate (avgdatact(nctrec(k),nspout(k),one))
      endif

      return
      end

c-----------------------------------------------------------------------
      subroutine dealloc
c-----------------------------------------------------------------------
c
c --- CALAVE    Version: 7.0.0          Level: 140912           DEALLOC
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  De-Allocate arrays
c
c --- UPDATES:
c
c --- v1.0 (120702) to TNG-2.0.0 (140912)
c        1. Add CALPUFF dataset TNG-3.0
c
c --- INPUT:
c
c --- OUTPUT:
c
c --- DEALLOC called by: (main)
c --- DEALLOC calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use header
      use inpdata
      use avgdata

      implicit none


c --- Logical arrays for header
      deallocate (lcompr,lsamp,ldisc,lctsg,lspunit)

c --- Character arrays for header
      deallocate (cmodel,ver,level)
      deallocate (dataset,dataver,datamod)
      deallocate (pmap,datum,axbtz,daten,utmhem)
      deallocate (clat0,clon0,clat1,clat2)
      deallocate (title,csout,csunit)
      deallocate (csrcnam,argrpnam)

c --- Integer arrays for header
      deallocate (ncom,ibyr,ibjul,ibhr,ibsec,irlg)
      deallocate (iavg,nsecdt,nxm,nym,nz,nssta)
      deallocate (ibcomp,iecomp,jbcomp,jecomp)
      deallocate (ibsamp,jbsamp,iesamp,jesamp)
      deallocate (meshdn,msource)
      deallocate (nsrctype,nsbytype,nrgrp,irgrp)
      deallocate (ndrec,nctrec,nspout,i2dmet,iutmzn)
      deallocate (ihill)
      deallocate (nsrc,ngx,ngy)
      
c --- Real arrays for header
      deallocate (xbtz,dxkm,dykm,xorigkm,yorigkm)
      deallocate (feast,fnorth)
      deallocate (rnlat0,relon0,xlat1,xlat2)
      deallocate (xrec,yrec,zrec,htrec)
      deallocate (xctr,yctr,zctr)



c --- Character arrays for inpdata
      deallocate (sname,cspecg,cspecd,cspecct)

c --- Integer arrays for inpdata
      deallocate (istype,isnum)
      deallocate (nyrb,njulb,nhrb,nsecb)
      deallocate (nyre,njule,nhre,nsece)

c --- Real arrays for inpdata
      deallocate (sxkm,sykm)
      deallocate (datag,datad,datact)

c --- Real work arrays for inpdata
      deallocate (workg,workd,workct)



c --- Real arrays for avgdata
      deallocate (avgdatag,avgdatad,avgdatact)

      return
      end

c-----------------------------------------------------------------------
      subroutine control(ictrl,iolst,cversion,clevel)
c-----------------------------------------------------------------------
c
c --- CALAVE    Version: 7.0.0          Level: 120915           CONTROL
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Obtain program control information
c
c --- UPDATES:
c --- v1.0 (120702) to v1.1 (120915)
c        1. Change all file pathname variables from C*132 to C*256
c           and make all line-input strings C*320
c
c --- INPUT:
c               ICTRL - integer   - Unit number of control file
c               IOLST - integer   - Unit number of output list file
c
c --- OUTPUT:
c     Module /USERINF/
c             nfiles,inpfile(),outfile(),lstfile,out_ext
c             start_hhmm,avgpd_hh,avgpd_mm,mode
c             start_hour,start_minute ,start_seconds
c             lcfiles
c
c --- CONTROL called by: (main)
c --- CONTROL calls:     DELIM, MK_CASE, RM_BLANK
c----------------------------------------------------------------------
c --- Access module data
      use userinf

      implicit none

c --- Declare arguments
      integer :: ictrl,iolst
      character(len=*)  ::  cversion,clevel

c --- Locals
      logical :: lprocess
      integer :: ihhadd
      integer :: i,ieof,ierr, id1,id2,ieq, icase, num
      character(len=320)  ::  blank320,aline320
      character(len=5)    ::  ulcase(2)
      character(len=1)    ::  dlim

      data dlim/'!'/
      data ulcase/'UPPER','LOWER'/

c --- Establish initial values (others in module userinf)
      do i=1,320
         blank320(i:i)=' '
      enddo
      lstfile=blank320(1:256)
      lstfile(1:10)='calave.lst'


c --- Scan control file for the list-file name, and
c --- number of input files (must allocate storage arrays)
c -----------------------------------------------------------
c --- (List file is not opened yet so error-writes will go to
c --- a FORT-file)
      nfiles=0
      ieof=0
      do while (ieof.EQ.0)
         aline320=blank320
         read(ictrl,'(a320)',iostat=ieof) aline320
         call DELIM(dlim,aline320,id1,id2,ieq,lprocess)
         if(lprocess) then
c ---       Convert to upper case
            call MK_CASE(iolst,'UPPER',aline320,ierr)
            if(ierr.NE.0) stop 'HALTED in CONTROL'
c ---       Interpret subset of inputs
            if(INDEX(aline320,'INPFILE').GT.0) then
               nfiles=nfiles+1
            elseif(INDEX(aline320,'LSTFILE').GT.0) then
               num=(id2-1)-(ieq+1)+1
               if(num.GT.256) then
                  write(*,*)'Error in CONTROL: file pathname '//
     &                      'exceeds 256 characters' 
                  write(*,*) TRIM(aline320)
                  stop
               endif
               lstfile=blank320(1:256)
               lstfile=aline320(ieq+1:id2-1)
c ---          Left-justify
               lstfile=ADJUSTL(lstfile)
            elseif(INDEX(aline320,'LCFILES').GT.0) then
               if(INDEX(aline320(ieq+1:id2-1),'T').GT.0) then
                  LCFILES=.TRUE.
               else
                  LCFILES=.FALSE.
               endif
            endif
         endif
      enddo

c --- Set pointer for ULCASE (1=upper, 2=lower)
      icase=1
      if(lcfiles) icase=2

c --- Open list file and write header
      call MK_CASE(iolst,ulcase(icase),lstfile,ierr)
      if(ierr.NE.0) stop 'HALTED in CONTROL'
      open(iolst,file=lstfile,status='unknown')
      write(iolst,*)
      write(iolst,*)
      write(iolst,*)'                    '//
     &              '      CALAVE MESSAGE SUMMARY'
      write(iolst,*)'                   VERSION:'//cversion//
     &              '   LEVEL:'//clevel
      write(iolst,*)
      write(iolst,*)

c --- Initial QA
c --- Must have 1 or more CALPUFF files
      if(nfiles.EQ.0) then
         write(iolst,*)'No CALPUFF file names are found in control file'
         write(iolst,*)'Make sure that names are assigned to INPFILE '
         stop 'HALTED in CONTROL -- see list file'
      endif

c --- Rewind control file to complete the processing
      REWIND(ictrl)

c --- Allocate arrays for storing input and output file names
      ALLOCATE (inpfile(nfiles),outfile(nfiles))

c --- Process control file again for all contents
c ------------------------------------------------
      nfiles=0
      ieof=0
      do while (ieof.EQ.0)
         aline320=blank320
         read(ictrl,'(a320)',iostat=ieof) aline320
         call DELIM(dlim,aline320,id1,id2,ieq,lprocess)
         if(lprocess) then
c ---       Convert to upper case
            call MK_CASE(iolst,'UPPER',aline320,ierr)
            if(ierr.NE.0) stop 'HALTED in CONTROL -- see list file'
c ---       Check list of expected inputs
c ---       -----------------------------
            if(INDEX(aline320,'INPFILE').GT.0) then
               num=(id2-1)-(ieq+1)+1
               if(num.GT.256) then
                  write(*,*)'Error in CONTROL: file pathname '//
     &                      'exceeds 256 characters' 
                  write(*,*) TRIM(aline320)
                  stop
               endif
               nfiles=nfiles+1
               inpfile(nfiles)=blank320(1:256)
               inpfile(nfiles)=aline320(ieq+1:id2-1)
               inpfile(nfiles)=ADJUSTL(inpfile(nfiles))
            elseif(INDEX(aline320,'LSTFILE').GT.0) then
c ---          Nothing:  already processed
            elseif(INDEX(aline320,'LCFILES').GT.0) then
c ---          Nothing:  already processed
            else
c ---          Remove blanks and reset pointers for rest
               call RM_BLANK(aline320)
               call DELIM(dlim,aline320,id1,id2,ieq,lprocess)
               if(INDEX(aline320,'AVGPD_HH').GT.0) then
                  read(aline320(ieq+1:id2-1),'(i)') avgpd_hh
               elseif(INDEX(aline320,'AVGPD_MM').GT.0) then
                  read(aline320(ieq+1:id2-1),'(i)') avgpd_mm
               elseif(INDEX(aline320,'START_HHMM').GT.0) then
                  start_hhmm=aline320(ieq+1:id2-1)
                  read(start_hhmm(1:2),'(i2.2)') start_hour
                  read(start_hhmm(3:4),'(i2.2)') start_minute
                  start_second=start_minute*60
               elseif(INDEX(aline320,'MODE').GT.0) then
                  read(aline320(ieq+1:id2-1),'(i)') mode
               elseif(INDEX(aline320,'OUT_EXT').GT.0) then
                  out_ext=blank320(1:40)
                  out_ext=aline320(ieq+1:id2-1)
               else
                  write(iolst,*)'Error in CALAVE control file'
                  write(iolst,*)'Assignment variable unknown:'
                  write(iolst,*)TRIM(aline320)
                  stop 'HALTED in CONTROL -- see list file'
               endif
            endif
         endif
      enddo

c --- QA
c --- Averaging period cannot be zero
      if(avgpd_hh.EQ.0 .AND. avgpd_mm.EQ.0) then
         write(iolst,*)'Error in CALAVE control file'
         write(iolst,*)'Averaging period cannot be zero'
         write(iolst,*)'Averaging Period   (hours)  ',avgpd_hh
         write(iolst,*)'Averaging Period (minutes)  ',avgpd_mm
         ierr=1
      endif
c --- Averaging period cannot be zero
      if(avgpd_hh.LT.0 .OR. avgpd_mm.LT.0) then
         write(iolst,*)'Error in CALAVE control file'
         write(iolst,*)'Averaging period cannot be negative'
         write(iolst,*)'Averaging Period   (hours)  ',avgpd_hh
         write(iolst,*)'Averaging Period (minutes)  ',avgpd_mm
         ierr=1
      endif
c --- Mode is 1 or 2
      if(mode.LT.1 .OR. mode.GT.2) then
         write(iolst,*)'Error in CALAVE control file'
         write(iolst,*)'Expected averaging mode = 1,2'
         write(iolst,*)'Found averaging mode    = ',mode
         ierr=1
      endif
c --- Start between 0000 and 2359
      if(start_hour.GT.23 .OR. start_minute.GT.59) then
         write(iolst,*)'Error in CALAVE control file'
         write(iolst,*)'Expected start time between 0000 and 2359'
         write(iolst,*)'Found start hour,minute = ',start_hour,
     &                  start_minute
         ierr=1
      endif
c --- End it?
      if(ierr.GT.0)then
         stop 'HALTED in CONTROL -- see list file'
      endif

c --- Condition averaging period
      if(avgpd_mm.GT.60) then
         ihhadd=avgpd_mm/60
         avgpd_mm=avgpd_mm-ihhadd*60
         avgpd_hh=avgpd_hh+ihhadd
      endif

c --- Construct output filenames
      do i=1,nfiles
         outfile(i)=blank320(1:256)
c ---    Remove any path from name (outputs go to current folder)
         id1=MAX(INDEX(inpfile(i),'/',back=.TRUE.),
     &           INDEX(inpfile(i),'\',back=.TRUE.))+1
         id2=LEN_TRIM(inpfile(i))
         outfile(i)=inpfile(i)(id1:id2)
         outfile(i)=ADJUSTL(outfile(i))
c ---    Add new extension
         outfile(i)=TRIM(outfile(i))//TRIM(out_ext)
      enddo

c --- Apply CASE choice to input and output filenames
      do i=1,nfiles
         call MK_CASE(iolst,ulcase(icase),inpfile(i),ierr)
         call MK_CASE(iolst,ulcase(icase),outfile(i),ierr)
      enddo
      if(ierr.NE.0) stop 'HALTED in CONTROL'

c --- Write interpreted control inputs to list file
      write(iolst,*)
      write(iolst,*)'CALAVE Control File Inputs'
      write(iolst,*)'--------------------------'
      write(iolst,*)
      write(iolst,*)'Averaging Period   (hours)  ',avgpd_hh
      write(iolst,*)'Averaging Period (minutes)  ',avgpd_mm
      write(iolst,*)'Mode  (1=Running  2=Block)  ',mode
      write(iolst,*)'Starting hour         (HH)  ',start_hour
      write(iolst,*)'Starting minute       (MM)  ',start_minute
      write(iolst,*)
      write(iolst,*)

      return
      end

c-----------------------------------------------------------------------
      subroutine comp(iolst,ioinp,ioout,ierr)
c-----------------------------------------------------------------------
c
c --- CALAVE    Version: 7.0.0          Level: 120702              COMP
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Process data records to the requested averaging period
c               and write averages to output file (same format);
c               Header of output file is already in place
c
c --- INPUT:
c          IOLST - integer - File unit number for output List file
c          IOINP - integer - File unit number for input CALPUFF file
c          IOOUT - integer - File unit number for output CALPUFF file
c
c --- OUTPUT:
c           IERR - integer - Error flag (0 = No Error)
c
c --- COMP called by: (main)
c --- COMP calls:     GETPER, MKAVG, PUTAVG
c----------------------------------------------------------------------
c --- Access module data
      use userinf
      use header
      use inpdata
      use avgdata

      implicit none

c --- Declare arguments
      integer :: iolst,ioinp,ioout,ierr

c --- Locals
      integer :: k,ndimx,ndimy,ndimd,ndimct
      integer :: i,inew,iold,ieof
      logical :: lrun

c --- All processing uses just 1 CALPUFF file here
      k=1

c --- Array allocation for k=1 should have resulted in exact match to
c --- shape actually used for the number of receptors --- verify this
      if(lsamp(k)) then
         ndimx=SIZE(datag,1)
         ndimy=SIZE(datag,2)
         if(ndimx.NE.ngx(k) .OR. ndimy.NE.ngy(k)) then
            write(iolst,*)'ERROR in COMP -- array allocation '//
     &                    'does not match gridded receptors in file'
            write(iolst,*)'   Allocated size: ',ndimx,ndimy
            write(iolst,*)'   Size in file  : ',ngx(k),ngy(k)
            ierr=1
            return
         endif
      endif
      if(ldisc(k)) then
         ndimd=SIZE(datad,1)
         if(ndimd.NE.ndrec(k)) then
            write(iolst,*)'ERROR in COMP -- array allocation '//
     &                    'does not match discrete receptors in file'
            write(iolst,*)'   Allocated size: ',ndimd
            write(iolst,*)'   Size in file  : ',ndrec(k)
            ierr=1
            return
         endif
      endif
      if(lctsg(k)) then
         ndimct=SIZE(datact,1)
         if(ndimct.NE.nctrec(k)) then
            write(iolst,*)'ERROR in COMP -- array allocation '//
     &                    'does not match CTSG receptors in file'
            write(iolst,*)'   Allocated size: ',ndimct
            write(iolst,*)'   Size in file  : ',nctrec(k)
            ierr=1
            return
         endif
      endif

c-----------------------------------------------------------------------
c --- Note:  There are navgpd model outputs within the requested
c ---        averaging period.  Model data for these are stored in
c ---        arrays of the same dimension.  For the first average,
c ---        the oldest modeled values are in array element 1 and the
c ---        newest is in array element navgpd.
c ---        For running averages, the next average is obtained by
c ---        replacing the oldest data with new data, so that the
c ---        oldest modeled values are now in array element 2 and the
c ---        newest is in array element 1.
c ---        For block averages, the next average is obtained by
c ---        replacing all of the data with new data, so that the
c ---        oldest modeled values remain in array element 1 and the
c ---        newest remains in array element navgpd.
c ---        If navgpd=1, running average is same as block average.
c-----------------------------------------------------------------------

c --- Set use of running or block average logic
      if(mode.EQ.1 .AND. navgpd.GT.1) then
         lrun=.TRUE.
      else
         lrun=.FALSE.
      endif

c --- Skip to start of processing period
10    call GETPER(iolst,ioinp,1,ierr,ieof)
      if(ierr.NE.0 .OR. ieof.NE.0) return
      if(start_hour.NE.nhrb(1,1) .OR.
     &   start_second.NE.nsecb(1,1)) goto 10

      if(lrun) then
c ---    Process running averages
c ---    Fill data storage arrays for first set of times to average, less 1
         do i=2,navgpd-1
            call GETPER(iolst,ioinp,i,ierr,ieof)
            if(ierr.NE.0 .OR. ieof.NE.0) return
         enddo

c ---    Set initial pointers to oldest and youngest times (to be updated)
         iold=navgpd
         inew=navgpd-1

c ---    Process all periods
         do while (ieof.EQ.0)

c ---       Update pointers to youngest and oldest times (current average)
            iold=iold+1
            if(iold.GT.navgpd) iold=iold-navgpd
            inew=inew+1
            if(inew.GT.navgpd) inew=inew-navgpd

c ---       Store new data in youngest slot
            call GETPER(iolst,ioinp,inew,ierr,ieof)
            if(ierr.NE.0 .OR. ieof.NE.0) return
c ---       Compute and store averages
            call MKAVG
c ---       Write averages to output file
            call PUTAVG(iolst,ioout,iold,inew,ierr)
            if(ierr.NE.0) return

         enddo

      else
c ---    Process block averages
c ---    Fill data storage arrays for first set of times to average
         do i=2,navgpd
            call GETPER(iolst,ioinp,i,ierr,ieof)
            if(ierr.NE.0 .OR. ieof.NE.0) return
         enddo

c ---    Set pointers to oldest and youngest times (to be updated)
         iold=1
         inew=navgpd
c ---    Compute and store averages
         call MKAVG
c ---    Write averages to output file
         call PUTAVG(iolst,ioout,iold,inew,ierr)
         if(ierr.NE.0) return

c ---    Process all remaining periods
         do while (ieof.EQ.0)

c ---       Replace all data
            do i=1,navgpd
               call GETPER(iolst,ioinp,i,ierr,ieof)
               if(ierr.NE.0 .OR. ieof.NE.0) return
            enddo
c ---       Compute and store averages
            call MKAVG
c ---       Write averages to output file
            call PUTAVG(iolst,ioout,iold,inew,ierr)
            if(ierr.NE.0) return

         enddo

      endif

      return
      end

c-----------------------------------------------------------------------
      subroutine getper(iolst,ioinp,it,ierr,ieof)
c-----------------------------------------------------------------------
c
c --- CALAVE    Version: 7.0.0          Level: 120702            GETPER
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Read and store model data records for 1 time period
c
c               Receptor arrays must be dimensioned EXACTLY for the
c               number(s) of receptors in the CALPUFF file!
c
c --- INPUT:
c          IOLST - integer - File unit number for output List file
c          IOINP - integer - File unit number for input CALPUFF file
c             IT - integer - Current time-period index
c
c --- OUTPUT:
c           IERR - integer - Error flag (0 = No Error)
c           IEOF - integer - End of file flag
c
c --- GETPER called by: COMP
c --- GETPER calls:     CMPRSIO, RDBDAT
c----------------------------------------------------------------------
c --- Access module data
      use header
      use inpdata

      implicit none

c --- Declare arguments
      integer :: iolst,ioinp,it,ierr,ieof

c --- Locals
      integer :: k,is,nsources,isp
      integer :: nxy,ncmprs,ione

c --- Initializations
      ione=0
      ieof=0
      ierr=0

c --- All processing uses just 1 CALPUFF file here
      k=1

c --- Source contribution configuration
      if(msource(k).EQ.0) then
c ---    Sum of all sources
         nsources=1
      else
c ---    Individual sources plus total
         nsources=nsrc(k)+1
      endif

c --- Number of gridded receptors
      nxy=ngx(k)*ngy(k)

c --- Source loop
c ---------------
      do is=1,nsources

c ---    Record 1:  Time period
         read(ioinp,end=999)
     &        nyrb(it,k),njulb(it,k),nhrb(it,k),nsecb(it,k),
     &        nyre(it,k),njule(it,k),nhre(it,k),nsece(it,k)

c ---    Record 2:  Source information
         read(ioinp,end=999)
     &        istype(is,k),isnum(is,k),sname(is,k),
     &        sxkm(is,k),sykm(is,k)

c ---    Output species-level loop
c --------------------------------
         do isp=1,nspout(k)
            if(lsamp(k)) then
               if(lcompr(k)) then
                  read(ioinp) ncmprs
                  call CMPRSIO(iolst,ioinp,'GETR',
     &                         datag(1,1,isp,is,it,k),nxy,
     &                         workg,ncmprs,cspecg(isp,k),ierr)
                  if(ierr.GT.0) return
               else
                  call RDBDAT(ioinp,cspecg(isp,k),
     &                        datag(1,1,isp,is,it,k),ngx(k),ngy(k))
               endif
            endif

            if(ldisc(k)) then
               if(lcompr(k))then
                  read(ioinp) ncmprs
                  call CMPRSIO(iolst,ioinp,'GETR',
     &                         datad(1,isp,is,it,k),ndrec(k),
     &                         workd,ncmprs,cspecd(isp,k),ierr)
                  if(ierr.GT.0) return
               else
                  call RDBDAT(ioinp,cspecd(isp,k),
     &                        datad(1,isp,is,it,k),ndrec(k),ione)
               endif
            endif

            if(lctsg(k)) then
               if(lcompr(k))then
                  read(ioinp) ncmprs
                  call CMPRSIO(iolst,ioinp,'GETR',
     &                         datact(1,isp,is,it,k),nctrec(k),
     &                         workct,ncmprs,cspecct(isp,k),ierr)
                  if(ierr.GT.0) return
               else
                  call RDBDAT(ioinp,cspecct(isp,k),
     &                        datact(1,isp,is,it,k),nctrec(k),ione)
               endif
            endif
         enddo
c --------------------------------
c ---    End output species-level loop

      enddo
c -------------------
c --- End source loop

      return

999   write(*,*)'Reached end of CALPUFF file ...'
      ieof=1
      return

      end

c-----------------------------------------------------------------------
      subroutine putavg(iolst,ioout,iold,inew,ierr)
c-----------------------------------------------------------------------
c
c --- CALAVE    Version: 7.0.0          Level: 120702            PUTAVG
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Write averaged data records for 1 time period
c
c               Receptor arrays must be dimensioned EXACTLY for the
c               number(s) of receptors in the CALPUFF file!
c
c --- INPUT:
c          IOLST - integer - File unit number for output List file
c          IOOUT - integer - File unit number for output CALPUFF file
c           IOLD - integer - Index for oldest period in data array
c           INEW - integer - Index for newest period in data array
c
c --- OUTPUT:
c           IERR - integer - Error flag (0 = No Error)
c
c --- PUTAVG called by: COMP
c --- PUTAVG calls:     CMPRSIO, WRBDAT
c----------------------------------------------------------------------
c --- Access module data
      use header
      use avgdata
      use inpdata

      implicit none

c --- Declare arguments
      integer :: iolst,ioout,iold,inew,ierr

c --- Locals
      integer :: k,is,nsources,isp
      integer :: nxy,ncmprs,ione
      integer :: kyrb,kjulb,khrb,ksecb,kyre,kjule,khre,ksece

c --- Initializations

c --- All processing uses just 1 CALPUFF file here
      k=1

c --- Source contribution configuration
      if(msource(k).EQ.0) then
c ---    Sum of all sources
         nsources=1
      else
c ---    Individual sources plus total
         nsources=nsrc(k)+1
      endif

c --- Number of gridded receptors
      nxy=ngx(k)*ngy(k)

c --- Start time for average (oldest period in data array)
      kyrb=nyrb(iold,k)
      kjulb=njulb(iold,k)
      khrb=nhrb(iold,k)
      ksecb=nsecb(iold,k)

c --- End time for average (newest period in data array)
      kyre=nyre(inew,k)
      kjule=njule(inew,k)
      khre=nhre(inew,k)
      ksece=nsece(inew,k)

c --- Source loop
c ---------------
      do is=1,nsources

c ---    Record 1:  Time period
         write(ioout)
     &         kyrb,kjulb,khrb,ksecb,kyre,kjule,khre,ksece

c ---    Record 2:  Source information
         write(ioout)
     &        istype(is,k),isnum(is,k),sname(is,k),
     &        sxkm(is,k),sykm(is,k)

c ---    Output species-level loop
c --------------------------------
         do isp=1,nspout(k)
            if(lsamp(k)) then
               if(lcompr(k)) then
                  ncmprs=nxy
                  call CMPRSIO(iolst,ioout,'PUTR',
     &                         avgdatag(1,1,isp,is),nxy,
     &                         workg,ncmprs,cspecg(isp,k),ierr)
                  if(ierr.GT.0) return
               else
                  call WRBDAT(ioout,cspecg(isp,k),
     &                        avgdatag(1,1,isp,is),ngx(k),ngy(k))
               endif
            endif

            if(ldisc(k)) then
               if(lcompr(k))then
                  ncmprs=ndrec(k)
                  call CMPRSIO(iolst,ioout,'PUTR',
     &                         avgdatad(1,isp,is),ndrec(k),
     &                         workd,ncmprs,cspecd(isp,k),ierr)
                  if(ierr.GT.0) return
               else
                  call WRBDAT(ioout,cspecd(isp,k),
     &                        avgdatad(1,isp,is),ndrec(k),ione)
               endif
            endif

            if(lctsg(k)) then
               if(lcompr(k))then
                  ncmprs=nctrec(k)
                  call CMPRSIO(iolst,ioout,'PUTR',
     &                         avgdatact(1,isp,is),nctrec(k),
     &                         workd,ncmprs,cspecct(isp,k),ierr)
                  if(ierr.GT.0) return
               else
                  call WRBDAT(ioout,cspecct(isp,k),
     &                        avgdatact(1,isp,is),nctrec(k),ione)
               endif
            endif
         enddo
c --------------------------------
c ---    End output species-level loop

      enddo
c -------------------
c --- End source loop

      return
      end

c-----------------------------------------------------------------------
      subroutine mkavg
c-----------------------------------------------------------------------
c
c --- CALAVE    Version: 7.0.0          Level: 120702             MKAVG
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Create averaged data records for 1 time period
c
c --- INPUT:
c
c --- OUTPUT:
c
c --- MKAVG called by: COMP
c --- MKAVG calls:     none
c----------------------------------------------------------------------
c --- Access module data
      use avgdata
      use inpdata
      use header
      use userinf

      implicit none

c --- Declare arguments
      integer :: iolst,ierr

c --- Locals
      integer :: k,nsources
      integer :: it,is,isp,i,j

c --- Initializations
      avgdatag=0.0
      avgdatad=0.0
      avgdatact=0.0

c --- All processing uses just 1 CALPUFF file here
      k=1

c --- Source contribution configuration
      if(msource(k).EQ.0) then
c ---    Sum of all sources
         nsources=1
      else
c ---    Individual sources plus total
         nsources=nsrc(k)+1
      endif

c --- Sum array data in order of the dimensions (outer to inner)

c --- Time periods
c     ------------
      do it=1,navgpd

c ---    Sources
c        -------
         do is=1,nsources

c ---       Output species-level
c           --------------------
            do isp=1,nspout(k)

c ---          Receptors
c              ---------
               if(lsamp(k)) then
                  do j=1,ngy(k)
                  do i=1,ngx(k)
                     avgdatag(i,j,isp,is)=avgdatag(i,j,isp,is)+
     &                                    datag(i,j,isp,is,it,k)
                  enddo
                  enddo
               endif
               if(ldisc(k)) then
                  do i=1,ndrec(k)
                     avgdatad(i,isp,is)=avgdatad(i,isp,is)+
     &                                  datad(i,isp,is,it,k)
                  enddo
               endif
               if(lctsg(k)) then
                  do i=1,nctrec(k)
                     avgdatact(i,isp,is)=avgdatact(i,isp,is)+
     &                                   datact(i,isp,is,it,k)
                  enddo
               endif

            enddo
         enddo
      enddo

c --- Average the sums by dividing by the number of periods
c --- Done if navgpd=1
      if(navgpd.EQ.1) return

c --- Sources
c     -------
      do is=1,nsources

c ---    Output species-level
c        --------------------
         do isp=1,nspout(k)

c ---       Receptors
c           ---------
            if(lsamp(k)) then
               do j=1,ngy(k)
               do i=1,ngx(k)
                  avgdatag(i,j,isp,is)=avgdatag(i,j,isp,is)/
     &                                 FLOAT(navgpd)
               enddo
               enddo
            endif
            if(ldisc(k)) then
               do i=1,ndrec(k)
                  avgdatad(i,isp,is)=avgdatad(i,isp,is)/
     &                               FLOAT(navgpd)
               enddo
            endif
            if(lctsg(k)) then
               do i=1,nctrec(k)
                  avgdatact(i,isp,is)=avgdatact(i,isp,is)/
     &                                FLOAT(navgpd)
               enddo
            endif

         enddo
      enddo

      return
      end
